@@ -1,4 +1,4 @@
-from typing import Optional, Tuple
+from typing import Optional, Tuple, Union
 import torch
 
 from pypaq.torchness.types import ACT, INI, TNS, DTNS
@@ -70,9 +70,9 @@ class LayBlockDRT(torch.nn.Module):
 
         self.res = LayRES(in_features=in_width, dropout=res_dropout) if residual else None
 
-    def forward(self, input:TNS) -> DTNS:
+    def forward(self, inp:TNS) -> DTNS:
 
-        out = self.ln_in(input)
+        out = self.ln_in(inp)
 
         out = self.denses[0](out)
         zsL = [zeroes(out)]
@@ -88,7 +88,7 @@ class LayBlockDRT(torch.nn.Module):
             out = self.drop_lay(out)
 
         if self.res:
-            out = self.res(input=out, bypass=input)
+            out = self.res(inp=out, bypass=inp)
 
         return {
             'out':  out,
@@ -153,11 +153,11 @@ class EncDRT(torch.nn.Module):
         for lix,lay in enumerate(self.drt_lays): self.add_module(f'lay_drt_{lix}',lay)
         if shared_lays and n_layers > 1: self.drt_lays *= n_layers
 
-    def forward(self, input:TNS) -> DTNS:
+    def forward(self, inp:TNS) -> DTNS:
 
         zsL = []
 
-        out = input
+        out = inp
 
         if self.in_drop_lay: # input dropout
             out = self.in_drop_lay(out)
@@ -183,6 +183,7 @@ class LayBlockCNN(torch.nn.Module):
     def __init__(
             self,
             n_filters: int,                             # num of filters
+            padded=                     True,           # if not padded reduces sequence length
             kernel_size: int=           3,              # layer kernel
             activation: ACT=            torch.nn.ReLU,  # global enc activation func
             lay_dropout: float=         0.0,
@@ -202,6 +203,7 @@ class LayBlockCNN(torch.nn.Module):
         super(LayBlockCNN, self).__init__()
 
         if kernel_size % 2 != 1: raise TorchnessException('LayBlockCNN kernel_size cannot be even number')
+        self.padded = padded
         self.kernel_size = kernel_size
 
         self.lay_ln = torch.nn.LayerNorm(
@@ -237,21 +239,21 @@ class LayBlockCNN(torch.nn.Module):
             dtype=          dtype,
             initializer=    initializer) if do_ldrt else None
 
-    def forward(self, input:TNS, history:Optional[TNS]=None) -> DTNS:
+    def forward(self, inp:TNS, history:Optional[TNS]=None) -> DTNS:
 
         zsL = []
+        out = self.lay_ln(inp)
 
-        out = self.lay_ln(input)
-
-        if history is None:
-            in_sh = list(input.shape)
-            pad_width = int((self.kernel_size-1)/2)
-            in_sh[-2] = pad_width
-            pad = torch.zeros(in_sh)
-            conc = [pad, out, pad] # pad both sides (encoder)
-        else:
-            conc = [history, out] # concatenate with history (casual encoder)
-        out = torch.concat(conc, dim=-2)
+        if self.padded:
+            if history is None:
+                in_sh = list(inp.shape)
+                pad_width = int((self.kernel_size-1)/2)
+                in_sh[-2] = pad_width
+                pad = torch.zeros(in_sh).to(out.device)
+                conc = [pad, out, pad] # pad both sides (encoder)
+            else:
+                conc = [history, out] # concatenate with history (casual encoder)
+            out = torch.concat(conc, dim=-2)
 
         out = self.lay_conv1D(out)
 
@@ -262,7 +264,9 @@ class LayBlockCNN(torch.nn.Module):
         if self.lay_drop:
             out = self.lay_drop(out)
 
-        out = self.res(input=out, bypass=input)
+        # it is not possible to do RES for not padded version
+        if self.padded:
+            out = self.res(inp=out, bypass=inp)
 
         if self.lay_DRT:
             lay_out = self.lay_DRT(out)
@@ -271,7 +275,7 @@ class LayBlockCNN(torch.nn.Module):
 
         return {
             'out':      out,
-            'state':    torch.split(input, split_size_or_sections=self.kernel_size-1, dim=-2)[-1],
+            'state':    torch.split(inp, split_size_or_sections=self.kernel_size-1, dim=-2)[-1],
             'zsL':      zsL}
 
 
@@ -286,6 +290,7 @@ class EncCNN(torch.nn.Module):
             # layer
             shared_lays: bool=          False,          # shared variables in enc_layers
             n_layers :int=              6,              # num of layers
+            padded=                     True,           # if not padded reduces sequence length
             kernel_size :int=           3,              # layer kernel
             n_filters :Optional[int]=   None,           # num of filters
             activation: ACT=            torch.nn.ReLU,  # global enc activation func
@@ -327,6 +332,7 @@ class EncCNN(torch.nn.Module):
 
         self.blocks = [LayBlockCNN(
             n_filters=          self.n_filters,
+            padded=             padded,
             kernel_size=        self.kernel_size,
             activation=         activation,
             lay_dropout=        lay_dropout,
@@ -351,26 +357,29 @@ class EncCNN(torch.nn.Module):
             dtype=              dtype)
 
     # prepares initial history for casual mode, history has shape [.., n_layers, kernel_size-1, n_filters]
-    def get_zero_history(self, input:TNS) -> TNS:
-        in_sh = list(input.shape)
+    def get_zero_history(self, inp:TNS) -> TNS:
+        in_sh = list(inp.shape)
         in_sh.insert(-2, self.n_layers)
         in_sh[-2] = self.kernel_size - 1
         in_sh[-1] = self.n_filters
-        return torch.zeros(in_sh)
+        return torch.zeros(in_sh).to(inp.device)
 
-    def forward(self, input:TNS, history:Optional[TNS]=None) -> DTNS:
+    def forward(
+            self,
+            inp: TNS,
+            history: Optional[TNS]= None) -> DTNS:
 
         states = []  # here we will store block states to concatenate them finally
         zsL = []
 
         if self.in_TFdrop_lay:
-            input = self.in_TFdrop_lay(input)
+            inp = self.in_TFdrop_lay(inp)
 
         if self.projection_lay:
-            input = self.projection_lay(input)
+            inp = self.projection_lay(inp)
 
-        output = input  # for 0 layers case
-        histories = torch.split(history,1,dim=-3) if history is not None else [None]*self.n_layers
+        output = inp  # for 0 layers case
+        histories = torch.split(history, 1, dim=-3) if history is not None else [None]*self.n_layers
         for block,hist in zip(self.blocks, histories):
             if hist is not None: hist = torch.squeeze(hist, dim=-3)
             block_out = block(output, history=hist)
@@ -464,12 +473,12 @@ class LayBlockTNS(torch.nn.Module):
 
     def forward(
             self,
-            src: TNS,
+            inp: TNS,
             task_query: Optional[TNS]=              None,           # forces task-attention mode (TAT)
-            src_mask: Optional[TNS]=                None,
-            src_key_padding_mask: Optional[TNS]=    None) -> DTNS:
+            inp_mask: Optional[TNS]=                None,
+            inp_key_padding_mask: Optional[TNS]=    None) -> DTNS:
 
-        x = src
+        x = inp
 
         if task_query is None: x = self.norm1(x) # norm first https://arxiv.org/pdf/2002.04745v1.pdf
         else: task_query = self.norm1(task_query) # LN on task_query
@@ -478,26 +487,27 @@ class LayBlockTNS(torch.nn.Module):
             query=              x if task_query is None else task_query,
             key=                x,
             value=              x,
-            key_padding_mask=   src_key_padding_mask,
+            key_padding_mask=   inp_key_padding_mask,
             need_weights=       False,
-            attn_mask=          src_mask)[0]
+            attn_mask=          inp_mask)[0]
 
         if self.dropout1:
             x = self.dropout1(x)
 
-        bypass = src if task_query is None else task_query
-        x = self.res1(input=x, bypass=bypass)
+        bypass = inp if task_query is None else task_query
+        x = self.res1(inp=x, bypass=bypass)
 
         return self.lay_drt(x)
 
 
-# Transformer Encoder (based on torch.nn.modules.transformer.TransformerEncoder)
+# Transformer Encoder (based on torch.nn.modules.transformer.TransformerEncoder), with TAT option and some others
 class EncTNS(torch.nn.Module):
 
     def __init__(
             self,
             num_layers: int=                        6,
             num_layers_TAT: int=                    0,
+            initial_TAT_avg: bool=                  True,   # how to prepare first block TAT task_query
             shared_lays: Optional[Tuple[int,...]]=  None,   # tuple defines layers groups with shared variables, e.g.: (2,2,2)
             max_seq_len: Optional[int]=             None,   # when given (int) adds positional embeddings (PE) to seq
             # block params
@@ -516,9 +526,20 @@ class EncTNS(torch.nn.Module):
         # positional embeddings (trainable)
         self.pos_emb = None
         if max_seq_len:
-            self.pos_emb = torch.nn.Parameter(torch.empty((max_seq_len, d_model), device=device, dtype=dtype))
+            self.pos_emb = torch.nn.Parameter(
+                data=   torch.empty(
+                    size=   (max_seq_len, d_model),
+                    device= device,
+                    dtype=  dtype))
             bert_initializer(self.pos_emb)
 
+        self.initial_task_query = None
+        if not initial_TAT_avg:
+            self.initial_task_query = torch.nn.Parameter(
+                data=           torch.empty(d_model, device=device, dtype=dtype),
+                requires_grad=  False) # TODO: check with experiments
+            bert_initializer(self.initial_task_query)
+
         # manage layers number
         num_layers_to_build = num_layers + num_layers_TAT
         if shared_lays is not None:
@@ -555,9 +576,10 @@ class EncTNS(torch.nn.Module):
             device=             device,
             dtype=              dtype)
 
-    def forward(self, src:TNS, mask:Optional[TNS]=None) -> DTNS:
+    # casual Transformer encoding
+    def _encode(self, inp:TNS, mask:Optional[TNS]=None) -> DTNS:
 
-        output = src
+        output = inp
 
         # add positional embeddings if needed
         if self.pos_emb is not None:
@@ -565,22 +587,64 @@ class EncTNS(torch.nn.Module):
 
         zsL = []
         for mod in self.layers:
-            block_out = mod(src=output, src_mask=mask)
+            block_out = mod(inp=output, inp_mask=mask)
             output = block_out['out']
             zsL += block_out['zsL']
 
         # pass through task-attention layers
         if self.layers_TAT:
+
             seq = output
-            task_query = torch.mean(seq, dim=-2, keepdim=True) # first initial reduce
+
+
+            if self.initial_task_query is not None:
+                task_query = self.initial_task_query                # from model
+                seq_shape = list(seq.shape)
+                seq_shape[-2] = 1
+                # TODO: not sure if it will work for batch of sequences, e.g. [128,64,512]
+                task_query = task_query.view(seq_shape)
+            else:
+                task_query = torch.mean(seq, dim=-2, keepdim=True)  # first initial from reduce
+
             for mod in self.layers_TAT:
-                block_out = mod(src=seq, task_query=task_query, src_mask=mask)
+                block_out = mod(inp=seq, task_query=task_query, inp_mask=mask)
                 task_query = block_out['out']
                 zsL += block_out['zsL']
-            output = task_query
+            output = torch.flatten(task_query,-2,-1) # remove seq (-2) dimension of size 1
 
         output = self.norm(output)
 
         return {
             'out':  output,
-            'zsL':  zsL}
\ No newline at end of file
+            'zsL':  zsL}
+
+    # pyramidal_encoding
+    def _encode_pyramidal(self, inp:TNS, pyramide: Union[Tuple[int],int]) -> DTNS:
+
+        if type(pyramide) is int: pyramide = (pyramide,)
+
+        inp = inp
+        zsL = []
+        for sl in pyramide:
+            in_split = torch.split(inp, sl, dim=0)
+            outL = [self._encode(inp=i) for i in in_split]
+            inp = [o['out'] for o in outL]
+            inp = torch.stack(inp, dim=0)
+            zsLL = [o['zsL'] for o in outL]
+            zsL += [e for el in zsLL for e in el]
+        out = self._encode(inp)
+        out['zsL'] += zsL
+        return out
+
+    def forward(
+            self,
+            inp: TNS,
+            mask: Optional[TNS]=                    None,
+            pyramide: Optional[Union[Tuple[int],int]]=  None) -> DTNS:
+
+        if pyramide:
+            if mask is not None:
+                raise TorchnessException('mask is not supported for pyramidal encoding')
+            return self._encode_pyramidal(inp, pyramide)
+        else:
+            return self._encode(inp, mask)
\ No newline at end of file
